/************************************************************
** $Id: test.c,v 0.3 2003/11/03 03:08:10 bcx Exp $
**
** Copyright Bryan Costales and Perforce Software, Inc. 2003
**
** This code is "open source" as defined by version 1.9 of
** the Open Source Definition from:
**
**	http://www.opensource.org/docs/definition.php.
**
** Tests run in this file can only be done before
** any threads are launched (i.e. from the main() routine).
** See the scripts in the tests directory for more details.
**
*************************************************************/

# define EXTERN extern
# include "slow.h"

/*
 * The following needed to be copied from milter.h in the
 * sendmail source for the routines in this file to work.
 */
# define MAX_MACROS_ENTRIES 4
PRIV Priv;
static struct smfi_str
{
        unsigned int    ctx_id;         /* thread id */
        unsigned int    ctx_sd;         /* socket descriptor */
        int             ctx_dbg;        /* debug level */
        time_t          ctx_timeout;    /* timeout */
        int             ctx_state;      /* state */
        smfiDesc_ptr    ctx_smfi;       /* filter description */
        unsigned long   ctx_pflags;     /* protocol flags */
        char            **ctx_mac_ptr[MAX_MACROS_ENTRIES];
        char            *ctx_mac_buf[MAX_MACROS_ENTRIES];
        char            *ctx_reply;     /* reply code */
        void            *ctx_privdata;  /* private data */
} Ctx;

void
test_init(int setpriv)
{
	(void) memset(&Ctx, '\0', sizeof(SMFICTX));
	Ctx.ctx_id = 1;
	if (setpriv == TRUE)
		Ctx.ctx_privdata = &Priv;
}

/*
** These various test routines all get a string for testing. 
** These are all passed using the -X command-line switch.
** Multilple arguments must be separated by commas, e.g. -X"a,b"
** All exit() after the individual test.
*/
void
test_connect(char *string)
{
	_SOCK_ADDR in;
	in_addr_t ip;
	char *host;
	char **argv;
	char buf[BUFSIZ];
	extern sfsistat xxfi_connect(SMFICTX *, char *, _SOCK_ADDR *);

	if (string == NULL)
	{
		(void) printf("Usage: -Dconnect -X \"hostname,ipaddr\"\n");
		exit(0);
	}

	argv = mkargv(string, ',');
	if (argv[0] == NULL || argv[1] == NULL)
	{
		(void) printf("Usage: -Dconnect -X \"hostname,ipaddr\"\n");
		exit(0);
	}

	if (((int)(ip = inet_network(argv[1]))) == -1)
	{
		(void) fprintf(stderr, "%s: %s\n", argv[1], strerror(errno));
		exit(0);
	}
	test_init(FALSE);
	LookBackOne = TRUE;
	/*
	 * First test with no macros defined and no AF_INET.
	 * Should get flags=[NET]
	 */
	in.sa_family = 0;
	(void)memcpy(in.sa_data, &ip, sizeof(PRIMARY_DATUM));
	host = strdup(argv[0]);
	(void) xxfi_connect(&Ctx, host, (_SOCK_ADDR *)&in);

	/*
	 * Second test with no macros defined and AF_INET defined.
	 * Should get empty flags=
	 */
	in.sa_family = AF_INET;
	(void)memcpy(in.sa_data, &ip, sizeof(PRIMARY_DATUM));
	host = strdup(argv[0]);
	(void) xxfi_connect(&Ctx, host, (_SOCK_ADDR *)&in);

	/*
	 * Then test the ${client_resolve} macro.
	 * Should get flags=[FOR]
	 */
	(void) sprintf(buf, "{client_resolve}, forged");
	Ctx.ctx_mac_ptr[0] = mkargv(buf, ',');
	Ctx.ctx_mac_buf[0] = buf;
	in.sa_family = AF_INET;
	(void)memcpy(in.sa_data, &ip, sizeof(PRIMARY_DATUM));
	host = strdup(argv[0]);
	(void) xxfi_connect(&Ctx, host, (_SOCK_ADDR *)&in);
	Ctx.ctx_mac_ptr[0] = NULL;
	Ctx.ctx_mac_buf[0] = NULL;

	/*
	 * Add the hostname to the list of MX hosts and
	 * try again.
	 * Should get <nil> for IP number, and flags=[MMX]
	 */
	MxHosts = push_list(argv[0], MxHosts);
	in.sa_family = AF_INET;
	(void)memcpy(in.sa_data, &ip, sizeof(PRIMARY_DATUM));
	host = strdup(argv[0]);
	(void) xxfi_connect(&Ctx, host, (_SOCK_ADDR *)&in);

	exit(0);
}

void
test_helo(char *string)
{
	char buf[BUFSIZ];
	extern sfsistat xxfi_helo(SMFICTX *, char *);

	test_init(TRUE);

	/*
	 * Test the ${illegalpipe} macro.
	 * Should get flags=[PIP]
	 */
	AdvanceWrite = TRUE;
	(void) sprintf(buf, "{illegalpipe}, true");
	Ctx.ctx_mac_ptr[0] = mkargv(buf, ',');
	Ctx.ctx_mac_buf[0] = buf;
	(void) xxfi_helo( &Ctx, string);
	Ctx.ctx_mac_ptr[0] = NULL;
	Ctx.ctx_mac_buf[0] = NULL;

	exit(0);
}

void
test_mailfrom(char *string)
{
	char **argv;
	extern sfsistat xxfi_envfrom(SMFICTX *, char **);

	if (string == NULL)
	{
		(void) printf("Usage: -Dmailfrom -X \"sender@domain\"\n");
		exit(0);
	}
	test_init(TRUE);
	CheckFromHeader = TRUE;

	/*
	 * Test the envelope recipient's host.
	 * Should get flags=[FRM] for a bad host name.
	 */
	argv = mkargv(string, ',');
	(void) xxfi_envfrom( &Ctx, argv);
	exit(0);
}

void
test_rcptto(char *string)
{
	char **argv;
	char *addr;
	extern sfsistat xxfi_envto(SMFICTX *, char **);

	if (string == NULL)
	{
		(void) printf("Usage: -Drcptto -X \"recipient@domain\"\n");
		exit(0);
	}
	test_init(TRUE);
	CheckFromHeader = TRUE;

	/*
	 * Should get flags=[HON] for a honey addr
	 */
	addr = strdup("bob@x.y.z.0.1.2.com");
	HoneyHosts = mkargv(addr, ',');
	(void) xxfi_envto(&Ctx, HoneyHosts);

	/*
	 * Test the envelope recipient's host.
	 * Three times to get the count to 3.
	 */
	argv = mkargv(string, ',');
	(void) xxfi_envto(&Ctx, argv);
	(void) xxfi_envto(&Ctx, argv);
	(void) xxfi_envto(&Ctx, argv);
	exit(0);
}

/*
 * Note that the .foo domain prevents these names from
 * being found in the actual MX list.
 */
static char *Received = "from spamhost.foo (bay2-f58.bay2.somehost.gov [65.54.247.58])\n\tby ourmx.foo (8.13.1/8.13.1) with ESMTP id h9VJ49rM005383\n\tfor <user@ourhost.foo>; Fri, 31 Oct 2003 11:04:09 -0800 (PST)";

void
test_header(char *string)
{
	char *addr;
	char *received;
	extern sfsistat xxfi_header(SMFICTX *, char *, char *);

	/* string is ignored */

	test_init(TRUE);

	/*
	 * Does it recognize the XTrackHeader?
	 */
	XTrackHeader = strdup("X-test-x-track");
	(void) xxfi_header(&Ctx, XTrackHeader, "testing");

	/*
	 * A Received: header not from one of our 
	 * MX servers, then one from one of our MX servers.
	 */
	LookBackOne = TRUE;
	Priv.ip = 0;
	received = strdup(Received);
	(void) xxfi_header(&Ctx, "Received", received);
	addr = strdup("ourmx.foo");
	MxHosts = push_list(addr, MxHosts);
	received = strdup(Received);
	(void) xxfi_header(&Ctx, "Received", received);

	/*
	 * A dozen vairations on recipient headers.
	 */
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("(John Doe) doe@ourhost.gov");
	(void) xxfi_header(&Ctx, "Cc", addr);
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("\"John Doe\" <doe@ourhost.gov>");
	(void) xxfi_header(&Ctx, "Cc", addr);
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("doe");
	(void) xxfi_header(&Ctx, "Bcc", addr);
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("doe@ourhost.gov (John Doe)");
	(void) xxfi_header(&Ctx, "Cc", addr);
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("<<<doe@ourhost.gov>>>");
	(void) xxfi_header(&Ctx, "Bcc", addr);
	Priv.hdrrcpts = free_list(Priv.hdrrcpts);
	addr = strdup("andy@ourhost.gov, <bob@ourhost.gov>, (Carl) carl@ourhost.gov");
	(void) xxfi_header(&Ctx, "To", addr);

	/*
	 * Next check the Message-ID header.
	 */
	addr = strdup("<a@b>");
	(void) xxfi_header(&Ctx, "Message-Id", addr);
	addr = strdup("<ab>");
	(void) xxfi_header(&Ctx, "Message-Id", addr);
	addr = strdup("<@b>");
	(void) xxfi_header(&Ctx, "Message-Id", addr);
	addr = strdup("<a@>");
	(void) xxfi_header(&Ctx, "Message-Id", addr);
	addr = strdup("a@b");
	(void) xxfi_header(&Ctx, "Message-Id", addr);

	/*
	 * Lastly check the From: header.
	 */
	CheckFromHeader = TRUE;
	addr = strdup("a");
	(void) xxfi_header(&Ctx, "From", addr);
	addr = strdup("a@");
	(void) xxfi_header(&Ctx, "From", addr);
	addr = strdup("a@host.foo");
	(void) xxfi_header(&Ctx, "From", addr);

	exit(0);
}


void
test_eoh(char *string)
{
	char **list;
	char buf[BUFSIZ];
	int  i;
	extern sfsistat xxfi_eoh(SMFICTX *);

	/* string is ignored */

	test_init(TRUE);

	CheckMessageId = TRUE;
	Priv.gotmid = FALSE;
	list = NULL;
	for (i = 0; i < 6; i++)
	{
		(void)sprintf(buf, "%d", i);
		list = push_list(buf, list);
	}
	Priv.hdrrcpts = list;
	(void) xxfi_eoh(&Ctx);
	exit(0);
}

void
test_close(char *string)
{
	extern sfsistat xxfi_close(SMFICTX *);

	/* string is ignored */

	test_init(TRUE);
	(void) xxfi_close(&Ctx);
	(void) printf("close: priv = %x\n", (int)Ctx.ctx_privdata);
	exit(0);
}

void
test_abort(char *string)
{
	extern sfsistat xxfi_abort(SMFICTX *);

	/* string is ignored */

	test_init(TRUE);
	(void) xxfi_abort(&Ctx);
	(void) printf("abort: flags=%s\n",
		bitset(BIT_MILTER_ABORTED, Priv.bits)?"[ABO]":"");     
	exit(0);
}

void
test_eom(char *string)
{
	char buf[BUFSIZ];
	extern sfsistat xxfi_eom(SMFICTX *);

	/* string is ignored */

	test_init(TRUE);
	(void) sprintf(buf, "{nbadrcpts}, 6");
	Ctx.ctx_mac_ptr[0] = mkargv(buf, ',');
	Ctx.ctx_mac_buf[0] = buf;
	(void) xxfi_eom(&Ctx);
	(void) printf("eom: badrcpts=%d\n", Priv.badrcpts);

	exit(0);
}

void
test_db(char *string)
{
	char **argv;
	char *pfname, *sfname;
	extern sfsistat xxfi_eom(SMFICTX *);

	if (string == NULL)
	{
		(void) printf("Usage: -Ddb -X \"primary.db, secondary.db\"\n");
		exit(0);
	}

	argv = mkargv(string, ',');
	if (argv[0] == NULL || argv[1] == NULL)
	{
		(void) printf("Usage: -Ddb -X \"primary.db, secondary.db\"\n");
		exit(0);
	}

	/*
	 * Use a path that will fail even if this is tested
	 * by root.
	 */
	pfname = "/foo/bar/foo/bar.db";
	errno = 0;
	Dp1 = open_database(pfname, TRUE);
	printf("db: %s: %s\n", pfname, strerror(errno));

	/*
	 * Using the file names specified, first
	 * open with no-force, then open with force.
	 */
	pfname = argv[0];
	errno = 0;
	Dp1 = open_database(pfname, FALSE);
	if (Dp1 == NULL)
		printf("db: %s: %s\n", pfname, strerror(errno));
	else
		close_database(Dp1);
		
	errno = 0;
	Dp1 = open_database(pfname, TRUE);
	if (Dp1 == NULL)
	{
		printf("db: %s: %s\n", pfname, strerror(errno));
		exit (0);
	}
	else
		printf("db: %s: successfully opened\n", pfname);

	sfname = argv[1];
	errno = 0;
	Dp2 = open_database(sfname, TRUE);
	if (Dp2 == NULL)
	{
		printf("db: %s: %s\n", sfname, strerror(errno));
		exit (0);
	}
	else
		printf("db: %s: successfully opened\n", sfname);

	/*
	 * Dummy up some records and add them
	 * to the database.
	 */
	Priv.iptxt = strdup("1.2.3.4");
	Priv.ip = inet_network(Priv.iptxt);
	Priv.bits = (BITMAP) 0;
	setbit(BIT_TOO_MANY_HDR_RCPTS, Priv.bits);
	setbit(BIT_TOO_MANY_ENV_RCPTS, Priv.bits);
	setbit(BIT_GOT_HONEYPOT, Priv.bits);
	setbit(BIT_BAD_MSG_ID, Priv.bits);
	setbit(BIT_BAD_FROM_HEAD, Priv.bits);
	setbit(BIT_BAD_HOST, Priv.bits);
	setbit(BIT_RBL_BAD, Priv.bits);
	setbit(BIT_MILTER_ABORTED, Priv.bits);
	setbit(BIT_ADVANCEWRITE, Priv.bits);
	setbit(BIT_FORGED, Priv.bits);
	setbit(BIT_FROMMX, Priv.bits);
	setbit(BIT_NOTINET, Priv.bits);
	Priv.envrcpts = 1234;
	Priv.badrcpts = 1234;
	Priv.honeyrcpts = 1234;
	Priv.nhdrrcpts = 1234;
	if (push_to_db(&Ctx, &Priv) != SMFIS_ACCEPT)
	{
		printf("%s: %s\n", Priv.iptxt, strerror(errno));
		exit(errno);
	}
	else
		printf("%s: successful write\n", Priv.iptxt);

	/*
	 * Now dump what we just wrote.
	 * There should just be one record.
	 */
	(void) dump_database(Dp1, Dp2, pfname, sfname, stdout);

	exit(0);
}
